:root {
    --color-black:        #141414;
    --color-dark-grey:    #292929;
    --color-neutral-grey: #808080;
    --color-light-grey:   #e1e1e1;
    --color-white:        #f5f5f5;
    --color-green:        #22c55e;
    --color-green-dim:    #1cab51;
    --color-blue:         #0ea5e9;
    --color-blue-dim:     #0b8fcb;
    --color-red:          #ef4444;
    --color-red-dim:      #d03a3a;
    --color-purple:       #a855f7;
    --color-purple-dim:   #9249d7;
    --color-yellow:       #eab308;
    --color-yellow-dim:   #cb9b06;
}

body {
    font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Helvetica, Arial, sans-serif;
    background-color: var(--color-black);
    color: var(--color-white);
    line-height: 1.0;
    display: flex;
    justify-content: center;
    align-items: center;
    min-height: 100svh;
    padding: 10px;
}
