# libmtmd.py
# https://github.com/ddh0/easy-llama/
# MIT License -- Copyright (c) 2024 Dylan Halladay

"""This file provides a Python interface to MTMD_API ("libmtmd"), which is originally defined
in `llama.cpp/tools/mtmd/mtmd.h`.

This file was last updated to match `llama.cpp/tools/mtmd/mtmd.h` as of this commit:
- Full SHA: `XXX`
- Commit message: `XXX`
- Date: XXX

This file's status with respect to the above commit is:
- XXX

Helpful references:
- `mtmd` README:
    [`llama.cpp/tools/mtmd/README.md`](https://github.com/ggml-org/llama.cpp/blob/master/tools/mtmd/README.md)
- `mtmd.h` at master:
    [`llama.cpp/tools/mtmd/mtmd.h](https://github.com/ggml-org/llama.cpp/blob/master/tools/mtmd/mtmd.h)"""

